package Nginx::Perl;

our $VERSION = '1.2.0.4';


1;
__END__

=head1 NAME

Nginx::Perl - full-featured perl support for nginx

=head1 DESCRIPTION

This is just a meta package to use as a dependency and 
for official package name on CPAN. 

Most of the documentation contained in L<Nginx>, check it out.

Web page: L<http://zzzcpan.github.com/nginx-perl>

=head1 SEE ALSO

L<http://zzzcpan.github.com/nginx-perl>,
L<Nginx>, L<Nginx::Test>, L<Nginx::Util>, L<Nginx::Redis>,
L<http://wiki.nginx.org/EmbeddedPerlModule>,
L<http://nginx.net/> 

=head1 AUTHOR

Alexandr Gomoliako <zzz@zzz.org.ua>

=head1 LICENSE

Copyright 2011 Alexandr Gomoliako. All rights reserved.

This module is free software. It may be used, redistributed and/or modified 
under the same terms as B<nginx> itself.

=cut

